/*
 *  Copyright (C) 2000 Marco Pesenti Gritti
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "galeon.h"

/** The pixmaps for the bookmarks */
GdkPixbuf *folder_pixmap = NULL;
GdkPixbuf *folder_open_pixmap = NULL;
GdkPixbuf *site_pixmap = NULL;

/** The root of the bookmark tree */
BookMarkItem *bookmarks_root = NULL;

/** The root of the temporary bookmark tree */
BookMarkItem *temp_bookmarks_root = NULL;

/* DnD */
const GtkTargetEntry bookmarks_dnd_targets [] = {
	{ "GALEON_BOOKMARK", 0, DND_TARGET_GALEON_BOOKMARK},
	{ "GALEON_URL", 0, DND_TARGET_GALEON_URL },
	{ "_NETSCAPE_URL", 0, DND_TARGET_NETSCAPE_URL},
	{ "text/uri-list", 0, DND_TARGET_TEXT_URI_LIST},
	{ "STRING", 0, DND_TARGET_STRING}
};
const gint bookmarks_dnd_targets_num_items = (
	sizeof (bookmarks_dnd_targets) / sizeof (GtkTargetEntry));

/* Private prototypes */
static BookMarkItem *bookmarks_xml_read_item (xmlNodePtr item);
static BookMarkItem *bookmarks_xml_read (xmlNodePtr item);
static void bookmarks_add_completion_recursively (BookMarkItem *b);
static void bookmarks_save_recursively (xmlDocPtr doc, xmlNodePtr root,
					BookMarkItem *b);
static BookMarkItem *bookmarks_find_by_nick (const gchar *nick,
					     BookMarkItem *root);


/**
 * bookmarks_load: load bookmarks from xml file
 */
void
bookmarks_load (void)
{
	gchar *bookfile;
	
	bookmarks_load_icons ();
	
	bookfile = g_concat_dir_and_file (g_get_home_dir (), 
					  "/.galeon/bookmarks.xml");
	
	bookmarks_root = bookmarks_load_from (bookfile);
	
	if (!bookmarks_root) {
		bookmarks_root = bookmarks_new_bookmark 
			(CATEGORY, _("Bookmarks"), NULL, NULL, NULL, NULL);
	}

	/* generate the autobookmarks category */
	autobookmarks_generate();
	
	bookmarks_add_completion_recursively (bookmarks_root);

	g_free (bookfile);
}

/**
 * bookmarks_load_icons: load the bookmarks icons
 */
void 
bookmarks_load_icons (void) 
{
	static gboolean done = FALSE;
	if (done) return;
	
	folder_pixmap = gdk_pixbuf_new_from_file (SHARE_DIR "/dir.xpm");
	
	folder_open_pixmap = gdk_pixbuf_new_from_file (SHARE_DIR "/dir_open.xpm");
			
	site_pixmap = gdk_pixbuf_new_from_file (SHARE_DIR "/i-bookmark.xpm");

	done = TRUE;
}

/**
 * bookmarks_load_from: Read a bookmarks tree
 * @file: The file to read, it should exist
 * Return value: The new allocated tree of bookmarks or NULL if error
 **/
BookMarkItem *
bookmarks_load_from (char *file)
{
	xmlDocPtr doc;
	xmlNodePtr item;
	BookMarkItem *b = NULL;

	if (!(g_file_exists (file))) {
		/* no bookmarks */
		return NULL;
	}

	doc = xmlParseFile (file);	
	
	if (doc) {
		item = doc->root;
		b = bookmarks_xml_read (item);
		xmlFreeDoc (doc);
		return b;
	} else {
		g_warning ("unable to parse bookmarks file: %s", file);
		return NULL;
	}
}

/**
 * bookmarks_xml_read: read all items from xml file
 */
static BookMarkItem * 
bookmarks_xml_read (xmlNodePtr item) 
{
	BookMarkItem *b, *result;
	item = item->childs;

	result = bookmarks_new_bookmark (CATEGORY, _("Bookmarks"),
					 NULL, NULL, NULL, NULL);
	
	while (item != NULL) {
		xmlNodePtr item2 = item;
		b = bookmarks_xml_read_item(item2);
		result->list = g_list_append(result->list, b);
		b->parent = result;
		item = item->next;
	}
	
	return result;
}

/**
 * bookmarks_xml_read_item: read an item from xml file
 */
static BookMarkItem *
bookmarks_xml_read_item (xmlNodePtr item)
{
	BookMarkItem *b;
	BookMarkItem *b2;
	gchar *ct, *expanded, *ccm;
	
	gchar *name = xmlGetPropISOLatin1 (item, "name");	
	gchar *notes = xmlGetProp (item, "notes");
	gchar *nick = xmlGetProp (item, "nick");
	gchar *pixmap_file = xmlGetProp (item, "pixmap");
  
	g_return_val_if_fail(item != NULL, NULL);

	if (!strcmp (item->name, "category")) {
		b = bookmarks_new_bookmark (CATEGORY, name, NULL, nick, notes,
					    pixmap_file);
		ct = xmlGetProp (item, "create_toolbar");
		ccm = xmlGetProp (item, "create_context_menu");
		expanded = xmlGetProp (item, "expanded");
		b->create_toolbar = (ct && !strcmp (ct, "TRUE")) ? TRUE : FALSE;
		b->create_context_menu = (ccm && !strcmp (ccm, "TRUE")) ? TRUE : FALSE;
		b->expanded = (expanded && !strcmp (expanded, "TRUE")) ? TRUE : FALSE;
		xmlFree (ct);
		xmlFree (ccm);
		xmlFree (expanded);
		item = item->childs;
		while (item != NULL) {
			xmlNodePtr item2 = item;
			b2 = bookmarks_xml_read_item (item2);
			b->list = g_list_append (b->list, b2);
			b2->parent = b;
			item = item->next;
		}
	} else if (!strcmp (item->name, "separator")) {
		b = bookmarks_new_bookmark (SEPARATOR, name, NULL, nick, notes,
					    pixmap_file);
	} else if (!strcmp (item->name, "autobookmarks")) {
		/* FIXME: remove code duplication */
		b = bookmarks_new_bookmark (AUTOBOOKMARKS, name, NULL, nick,
					    notes, pixmap_file);
		ct = xmlGetProp (item, "create_toolbar");
		ccm = xmlGetProp (item, "create_context_menu");
		expanded = xmlGetProp (item, "expanded");
		b->create_toolbar = (ct && !strcmp (ct, "TRUE")) ? TRUE : FALSE;
		b->create_context_menu = (ccm && !strcmp (ccm, "TRUE")) ? TRUE : FALSE;
		b->expanded = (expanded && !strcmp (expanded, "TRUE")) ? TRUE : FALSE;
		xmlFree (ct);
		xmlFree (ccm);
		xmlFree (expanded);
		autobookmarks_root = b;
	} else { /* site */
		gchar *url = xmlGetProp (item, "url");
		b = bookmarks_new_bookmark (SITE, name, url, nick, notes,
					    pixmap_file);
		ccm = xmlGetProp (item, "create_context_menu");
		b->create_context_menu = (ccm && !strcmp (ccm, "TRUE")) ? TRUE : FALSE;
		xmlFree (ccm);
		if (url) g_free (url);
	}
	
	if (name) g_free (name);
	if (notes) g_free (notes);
	if (nick) g_free (nick);
	if (pixmap_file) g_free (pixmap_file);
	
	return b;
}

void
temp_bookmarks_save (void)
{
	char *file = g_strconcat (g_get_home_dir(), "/.galeon/temp_bookmarks.xml", NULL);
	if (temp_bookmarks_root && temp_bookmarks_window 
	    && temp_bookmarks_window->dirty) {
		bookmarks_save_as (temp_bookmarks_root, file);
		temp_bookmarks_window->dirty = FALSE;
	}
	g_free (file);
}

/**
 * bookmarks_save: save all bookmarks to the xml file
 */
void
bookmarks_save (void) 
{
	bookmarks_update_menu ();
	bookmarks_update_tb ();
	if ((!bookmarks_editor) || (bookmarks_editor->dirty)) {
		char *savefile = g_strconcat (g_get_home_dir(), 
					      "/.galeon/bookmarks.xml", NULL);
		bookmarks_save_as (bookmarks_root, savefile);
		if (bookmarks_editor) bookmarks_editor->dirty = FALSE;
		g_free (savefile);
	}
}

/**
 * bookmarks_save_as: Save a set of bookmarks
 * @root: the root category to be saved
 * @file: the filename, if already exists will be overwriten
 **/
/* FIXME: check for errors (permisions...) */
void 
bookmarks_save_as (BookMarkItem *root, char *file)
{
	xmlNodePtr node;
	xmlDocPtr doc = xmlNewDoc ("1.0");
	GList *li;
	node = xmlNewDocNode (doc, NULL, "bookmarks", NULL);
	
  	for (li = root->list; li != NULL; li = li->next) {
		BookMarkItem *b = li->data;
		bookmarks_save_recursively (doc, node, b);
	}	

	xmlDocSetRootElement (doc, node);
	xmlSaveFile (file, doc);
	xmlFreeDoc (doc);
}

/**
 * bookmarks_save_recursively: recursively save a bookmarks tree to the xml file 
 */
static void
bookmarks_save_recursively (xmlDocPtr doc, xmlNodePtr root, BookMarkItem *b)
{
	BookMarkItem *b2;
	xmlNodePtr node;
	GList *li;
  
	switch (b->type) {
	case CATEGORY:
		node = xmlNewDocNode (doc, NULL, "category", NULL);
		xmlSetPropISOLatin1 (node, "name", b->name);
		xmlSetProp (node, "create_toolbar", 
			    b->create_toolbar ? "TRUE" : "FALSE");
		xmlSetProp (node, "create_context_menu", 
			    b->create_context_menu ? "TRUE" : "FALSE");
		xmlSetProp (node, "expanded", 
			    b->expanded ? "TRUE" : "FALSE");
		xmlSetProp(node, "notes", b->notes);
		if (b->pixmap_file)
			xmlSetProp(node, "pixmap", b->pixmap_file);
		xmlAddChild (root, node);
    
		for (li = b->list; li != NULL; li = li->next) {
			b2 = li->data;
			bookmarks_save_recursively(doc, node, b2);
		}
		break;

	case SEPARATOR:
		node = xmlNewDocNode(doc, NULL, "separator", NULL);
		xmlAddChild(root, node);
		break;

	case SITE:
		node = xmlNewDocNode(doc, NULL, "site", NULL);
		xmlSetPropISOLatin1 (node, "name", b->name);
		xmlSetProp(node, "url", b->url);    
		xmlSetProp(node, "nick", b->nick);
		xmlSetProp(node, "notes", b->notes);
		xmlSetProp (node, "create_context_menu", 
			    b->create_context_menu ? "TRUE" : "FALSE");
		if (b->pixmap_file)
			xmlSetProp(node, "pixmap", b->pixmap_file);
		xmlAddChild(root, node);
		break;

	case AUTOBOOKMARKS:
		node = xmlNewDocNode(doc, NULL, "autobookmarks", NULL);
		xmlSetPropISOLatin1 (node, "name", b->name);
		xmlSetProp (node, "create_toolbar", 
			    b->create_toolbar ? "TRUE" : "FALSE");
		xmlSetProp (node, "create_context_menu", 
			    b->create_context_menu ? "TRUE" : "FALSE");
		xmlSetProp (node, "expanded", 
			    b->expanded ? "TRUE" : "FALSE");
		xmlSetProp(node, "notes", b->notes);
		if (b->pixmap_file)
			xmlSetProp(node, "pixmap", b->pixmap_file);
		xmlAddChild(root, node);
		break;
	}
}

/**
 * bookmarks_remove_recursively: recursively remove a tree of bookmarks items
 * Doesn't update tree items
 */
void 
bookmarks_remove_recursively (BookMarkItem *b) {
	GList *position, *li;
  
	if (b->parent) {
		position = g_list_find (b->parent->list, b);
		b->parent->list = 
			g_list_remove_link (b->parent->list, position);
	}

	if (b->type == CATEGORY || b->type == AUTOBOOKMARKS)
		for (li = b->list; li != NULL; li = li->next) 
			bookmarks_remove_recursively (li->data);

	bookmarks_free_bookmark (b);
}

/**
 * bookmarks_new_bookmark: Creates a new allocated BookMarkItem
 * @type: the type of the bookmark
 * @name: the name, if NULL the url is used
 * @url: the url (ignored if type != SITE)
 * @nick: the nickname for the bookmark, if any
 * @notes: the comments aout the bookmark
 * 
 * This function allocates a BookMarkItem and initializes its fields to sane values.
 * Use it instead of g_new (BookMarkItem, 1) and manually initializing the fields. 
 * The returned BookMarkItem has it's parent set to NULL, so you have to add it to 
 * the bookmarks list yourself (if you need it)
 * All the strings passed to this function are g_strduped
 * 
 * Return value: The new BookMarkItem
 **/
BookMarkItem *
bookmarks_new_bookmark (BookMarkType type, const char *name, const char *url, 
			const char *nick, const char *notes, 
			const char *pixmap_file)
{
	BookMarkItem *bookmark;

	g_return_val_if_fail (!((type == SITE) && (url == NULL)), NULL);
	g_return_val_if_fail (!((type == CATEGORY) && (name == NULL)), NULL);

	bookmark = g_new0 (BookMarkItem, 1);
	bookmark->type = type;

	switch (bookmark->type) {
	case AUTOBOOKMARKS:
	case CATEGORY:
		bookmark->name = g_strdup (name);
		bookmark->url = NULL;
		break;
	case SITE:
		if (name) 
			bookmark->name = g_strdup (name);
		else
			bookmark->name = g_strdup (url);
		bookmark->url = g_strdup (url);
		break;
	default:
		bookmark->name = bookmark->url = NULL;
		break;
	}
	bookmark->list = NULL;
	bookmark->tree_item = NULL;
	bookmark->parent = NULL;
	bookmark->create_toolbar = FALSE;
	bookmark->create_context_menu = FALSE;
	if (nick) 
		bookmark->nick = g_strdup (nick);
	else
		bookmark->nick = g_strdup ("");
	if (notes)
		bookmark->notes = g_strdup (notes);
	else
		bookmark->notes = g_strdup ("");
	if (pixmap_file != NULL && strlen(pixmap_file) != 0)
	{
		bookmark->pixmap_file = g_strdup (pixmap_file);
		bookmark->pixmap = gdk_pixbuf_new_from_file (pixmap_file);
	}
	else
	{
		bookmark->pixmap_file = g_strdup ("");
		bookmark->pixmap = NULL;
	}
	bookmark->expanded = TRUE;
	return bookmark;
}

/**
 * bookmarks_copy_bookmark: copy a bookmark
 * @b: the bookmark
 * 
 * copy a bookmarks and its children
 * 
 * Return value: the new bookmark
 **/
BookMarkItem *
bookmarks_copy_bookmark (BookMarkItem *b)
{
	GList *l;
	BookMarkItem *new = bookmarks_new_bookmark (b->type, b->name, b->url,
						    b->nick, b->notes, 
						    b->pixmap_file);
	
	new->create_toolbar = b->create_toolbar;
	new->create_context_menu = b->create_context_menu;
	new->expanded = b->expanded;
	if (b->pixmap != NULL)
	{
		new->pixmap = gdk_pixbuf_copy (b->pixmap);
	}
	if (b->type == CATEGORY) {
		for (l = b->list; l != NULL; l = g_list_next (l)) {
			BookMarkItem *child = l->data;
			BookMarkItem *new_child = bookmarks_copy_bookmark (child);
			new_child->parent = new;
			new->list = g_list_append (new->list, new_child);
		}
	}
	return new;
}

/**
 * bookmarks_free_bookmark: free a bookmark
 * @b: the bookmark to be freed, if NULL thenb do nothing
 * 
 * Frees the bookmarks and all the strings referenced by it and the list of children 
 * if it's a CATEGORY (but does not free the children). Use it instead of g_free. 
 **/
void
bookmarks_free_bookmark (BookMarkItem *b)
{
	if (! b) return;

	if (b->name) g_free (b->name);
	if (b->url) g_free (b->url);
	if (b->nick) g_free (b->nick);
	if (b->pixmap_file) g_free (b->pixmap_file);
	if (b->pixmap) gdk_pixbuf_unref (b->pixmap);
	if (b->notes) g_free (b->notes);
	if (b->list) g_list_free (b->list);

	g_free (b);
}

/**
 * bookmarks_insert_bookmark: insert a bookmark
 * @b: the bookmark t insert
 * @near: a bookmark that should be near @b (a brother or its parent)
 **/
void
bookmarks_insert_bookmark (BookMarkItem *b, BookMarkItem *near,
			   GtkCListDragPos insert_pos)
{
	BookMarkItem *parent;
	GList *position_node;
	gint position;

	if (near == NULL)
		parent = bookmarks_root;
	else
	{
		if (insert_pos == GTK_CLIST_DRAG_INTO &&
		    (near->type == CATEGORY || near->type == AUTOBOOKMARKS))
			parent = near;
		else
			parent = near->parent;
	}

	g_assert (parent != NULL);
	b->parent = parent;

	position_node = g_list_find (parent->list, near);
	if (!position_node)
		parent->list = g_list_prepend (parent->list, b);
	else
	{
		position = g_list_position (parent->list, position_node);

		switch (insert_pos)
		{
		case GTK_CLIST_DRAG_NONE:
		case GTK_CLIST_DRAG_BEFORE:
			break;
		case GTK_CLIST_DRAG_INTO:
			if (near == NULL)
				break;
			if (near->type == SITE ||
			    near->type == SEPARATOR)
				position++;
			break;				
		case GTK_CLIST_DRAG_AFTER:
			position++;
			break;
		default:
			break;
		}
		parent->list = g_list_insert (parent->list, b, position);
	}
}

/**
 * bookmarks_move_bookmark: Moves a bookmark
 * @controls: the controls of the editor
 * @b: the bookmark to move
 * @where: if 0 then move up, else move down
 * Returns: 1 if bookmark is now first or last in the list, 0 otherwise
 *
 * Moves a bookmark up or down and updates both the edition ctree and the 
 * bookmarks structure 
 **/
gint
bookmarks_move_bookmark (bookmarks_editor_controls *controls, BookMarkItem *b, int where)
{
	BookMarkItem *parent, *grandparent, *other;
	GList *pos, *other_pos;
	
	parent = b->parent;
	if (!parent)
		return 1;

	grandparent = parent->parent;
	pos = g_list_find (parent->list, b);

	if (!where) /* up */
		other_pos = g_list_previous(pos);
	else /* down */
		other_pos = g_list_next(pos);
	
	if (other_pos) {
		other = other_pos->data;
		if (other->type == CATEGORY) {
			/* expand the category and move b into it */
			parent->list = g_list_remove_link (parent->list, pos);
			gtk_ctree_expand (GTK_CTREE (controls->ctree), other->tree_item);
			
			if (!where) { 
				/* up */
				other->list = g_list_append (other->list, b);
				b->parent = other;
			} else {
				/* down */
				b->parent = other;
				other->list = g_list_prepend (other->list, b);
			}
		} else {
			gint newindex = g_list_position (parent->list, other_pos);
			b->parent->list = g_list_remove_link (b->parent->list, pos);
			b->parent->list = g_list_insert (parent->list, b, newindex);
		}
	} else {
		/* move b to its grandparent*/
		if  (!grandparent)
			return 1;
		
		parent->list = g_list_remove_link(parent->list, pos);

		if (!where) {
			/* up */
			grandparent->list = g_list_insert 
				(grandparent->list, b, g_list_position 
				 (grandparent->list, g_list_find(grandparent->list, 
								 parent)));
			b->parent = grandparent;
		} else {
			/* down */
			GList *next = g_list_find(grandparent->list, parent)->next;
			grandparent->list = g_list_insert (grandparent->list, b, 
							   g_list_position 
							   (grandparent->list, next));
			b->parent = grandparent;
		}
	}
	bookmarks_editor_place_tree_item (controls, b);
	gtk_ctree_select ( GTK_CTREE (controls->ctree), b->tree_item);

	return 0;
}

/**
 * Renames a bookmark (wich has just been added) or deletes it if the user cancelled 
 */
void 
bookmarks_string_request_callback (gchar *string, BookMarkItem *b) 
{
	if (!string) { /* User canceled */
		bookmarks_remove_recursively (b);
	} else { 
		g_free (b->name);
		b->name = string;
		
		if (strlen (b->name) == 0) {
			g_free (b->name);
			b->name = g_strdup ( _("Untitled"));
		}

		if (b->url &&
		    gnome_config_get_bool (CONF_GENERAL_FAVICONS_ENABLED))
			get_siteicon (b->url);
		
		if (bookmarks_editor)
			bookmarks_editor_place_tree_item (bookmarks_editor, b);
		bookmarks_update_menu ();
		bookmarks_save (); 
	}
}

/**
 * add_temp_bookmark: Adds a temporal bookmark
 * @type: The type of the bookmark
 * @name: The name of the bookmark, if NULL or "" the url will be used instead.
 * @url: The url of the bookmark
 * @parent: The parent bookmark or NULL if the bookmark should be added to the 
 * main temp category
 * Return value: The new bookmark
 **/
BookMarkItem *
add_temp_bookmark (BookMarkType type, const gchar *name, const gchar *url, 
		   BookMarkItem *parent)
{
	BookMarkItem *bookmark;

	bookmark = bookmarks_new_bookmark (type, name, url, NULL, NULL, NULL);

	if (!temp_bookmarks_root) {
		temp_bookmarks_init ();
	}
	if (!parent) parent = temp_bookmarks_root;
	
	bookmark->parent = parent;
	parent->list = g_list_append (parent->list, bookmark);
	bookmarks_editor_place_tree_item (temp_bookmarks_window, bookmark);
	temp_bookmarks_window->dirty = TRUE;

	/* Save the temp bookmarks, for safety in case of a crash */
	temp_bookmarks_save ();

	return bookmark;
}

/**
 * bookmarks_list_all_urls:
 * -----------------------
 * get a list of all bookmarked URLs
 */
GSList *
bookmarks_list_all_urls (BookMarkItem *bookmark)
{
	GList *l;
	GSList * result;

	g_return_val_if_fail(bookmark != NULL, NULL);
	g_return_val_if_fail(bookmark->type == CATEGORY, NULL);

	result = g_slist_alloc();

	for (l = bookmark->list; l != NULL; l = l->next)
	{
		BookMarkItem *b = l->data;
		if (b->type == CATEGORY) {
			g_slist_concat(result, bookmarks_list_all_urls(b));
		} else
			if (b->type == SITE) {
				g_slist_append(result, b->url);
			}
	}
	return result;
}

/**
 * get a singly linked list of all the category/folder bookmarks 
 */
GSList * 
bookmarks_list_all_folders (BookMarkItem *bookmark)
{
	GList *l;
	GSList * result;

	g_return_val_if_fail(bookmark != NULL, NULL);
	g_return_val_if_fail(bookmark->type == CATEGORY, NULL);

	result = g_slist_alloc();

	for (l = bookmark->list; l != NULL; l = l->next)
	{
		BookMarkItem *b = l->data;
		if (b->type == CATEGORY) {
			g_slist_append(result, b);
			g_slist_concat(result, bookmarks_list_all_folders(b));
		}
	}
	return result;
}


/**
 * bookmarks_parse_nick: returns the url of a given nick with one arg
 * @text: the text to try to parse
 * @root: the root bookmark for start the search, if NULL the usual tree will be used
 * 
 * Tries to find a bookmarkitem wich has a nick like the first word of @text and if 
 * found return the URL result of substituting each %s in the BookmarkItem url by the 
 * text after the first " " in @text
 * 
 * Return value: the resulting URL or NULL if not matching nick found
 **/
gchar *
bookmarks_parse_nick (gchar *text, BookMarkItem *root)
{
	gchar *nick;
	gchar *args;
	gchar *url;
	gchar *space;
	BookMarkItem *b;

	/* remove any leading spaces */
	while (*text == ' ')
		text++;
	if (*text == '\0') 
		return NULL;

	/* find the nick */
	space = strstr (text, " ");
	if (space) 
	{
		/* split out nickname and arguments */
		nick = g_strndup (text, space - text);
 		args = g_strdup (space + 1);

		/* find bookmark */
		b = bookmarks_find_by_nick (nick, root);
		g_free (nick);

		/* check we found it */
		if (b == NULL)
			return NULL;

		/* do substitution */		
		url = bookmarks_substitute_argument (b, args);
		g_free (args);

		/* return completed url */
		return url;
	}
	else
	{
		/* the whole thing is the nick */
		nick = g_strdup (text);

		/* find bookmark */
		b = bookmarks_find_by_nick (nick, root);
		g_free (nick);

		/* check we found it */
		if (b == NULL)
			return NULL;

		/* return copy of the url */
		return g_strdup(b->url);
	}
}

/**
 * bookmarks_substitute_argument:
 **/
gchar *
bookmarks_substitute_argument (BookMarkItem *b, gchar *arguments)
{
	gchar *ret = NULL;
	gchar *s = b->url;
	int expected_args = 0;

	/* count %s in the URL */
	while ((s = strstr(s, "%s"))) {
	    ++expected_args;
	    ++s;
	}
	
	ret = g_strdup (b->url);

	if (expected_args <= 1) {
	    /*  do the substitution */
	    while ( (s = strstr (ret, "%s")) ) {
		gchar *new1, *new2;
		new1 = g_strndup (ret, s - ret);
		new2 = g_strconcat (new1, arguments, s + 2, NULL);
		g_free (new1);
		g_free (ret);
		ret = new2;
	    }
	}
	else {
	    /* try to substitute matching arguments */
	    gchar **args;
	    int arg = 0;
	    args = g_strsplit(arguments, " ", -1);
	    while ( (s = strstr (ret, "%s")) ) {
		gchar *new1, *new2;
		if (!args[arg])
		    break;
		new1 = g_strndup (ret, s - ret);
		new2 = g_strconcat (new1, args[arg++], s + 2, NULL);
		g_free (new1);
		g_free (ret);
		ret = new2;
	    }
	    g_strfreev(args);
	}

	return ret;
}

/**
 * bookmarks_find_by_nick: search for a BookMarkItem with a nick
 * @nick: the nick to search for.
 * @r: the root bookmark for start the search, if NULL the usual tree will be used
 * 
 * Search a bookmarks category recursivley for a bookmark with the given nick
 * 
 * Return value: the found BookMarkItem or NULL if nothing found
 **/
static BookMarkItem *
bookmarks_find_by_nick (const gchar *nick, BookMarkItem *r)
{
	GList *l;
	BookMarkItem *b;
	if (!r) r = bookmarks_root;

	if ((r->nick) && !strcmp (r->nick, nick)) return r;
	if (r->type == CATEGORY)
		for (l = r->list; l != NULL; l = g_list_next (l)) {
			b = bookmarks_find_by_nick (nick, l->data);
			if (b) return b;
		}
	return NULL;
}

/**
 * bookmarks_save_as: convert a bookmark to a string
 * @bookmark: the bookmark to be saved (can be a category)
 *
 * Return value: the bookmarks as a string.
 **/
gchar * 
bookmarks_item_to_string (BookMarkItem *bookmark)
{
	xmlNodePtr node;
	xmlDocPtr doc = xmlNewDoc ("1.0");
	xmlChar *mem;
	int size;

	node = xmlNewDocNode (doc, NULL, "single_bookmark", NULL);

	bookmarks_save_recursively (doc, node, bookmark);

	xmlDocSetRootElement (doc, node);

	xmlDocDumpMemory (doc, &mem, &size);

	xmlFreeDoc (doc);

	return mem;
}

/**
 * bookmarks_load_from_string: Read a bookmark from a string
 * @string: the string to parse.
 * Return value: The new allocated bookmark or NULL if error
 **/
BookMarkItem *
bookmarks_item_from_string (char *string)
{
	xmlDocPtr doc;
	xmlNodePtr item;
	BookMarkItem *b = NULL;

	doc = xmlParseMemory (string, strlen (string));	
	
	if (doc) {
		item = doc->root->childs;
		b = bookmarks_xml_read_item (item);
		xmlFreeDoc (doc);
	} else {
		g_warning ("Error parsing: %s", string);
	}
	return b;
}

/**
 * bookmarks_translate_string
 *   Converts non-alphanumeric characters in a smart bookmark string
 *   to the form %{hex key code}
 *   The returned string should be freed.
 **/
gchar *
bookmarks_translate_string (gchar *smart_bm_string)
{
	gchar *translated_string, *tstring_ptr;
	guchar curr_char;
	gchar key_code[4];

	/* each translated character will have the form %XX, so we'll
	   allocatedenough space for 3*current string length */
	translated_string = g_malloc0(3*strlen(smart_bm_string)+1);
	tstring_ptr = translated_string;

	while (*smart_bm_string)
	{
		curr_char = *(unsigned char *)smart_bm_string;
		if ( (curr_char >= 'a' && curr_char <= 'z') ||
		     (curr_char >= 'A' && curr_char <= 'Z') ||
		     (curr_char >= '0' && curr_char <= '9') )
			*tstring_ptr++ = curr_char;
		else if (curr_char == ' ')
			*tstring_ptr++ = '+';
		else
		{
			sprintf(key_code, "%%%X", curr_char);
			strcat(translated_string, key_code);
			tstring_ptr += 3;
		}
		smart_bm_string++;
	}

	return (translated_string);
}

/**
 * bookmarks_folder_open_all_items: Loads all bookmarks in a folder 
 **/
void
bookmarks_folder_open_all_items (GaleonEmbed *embed, BookMarkItem *bi,
				 gboolean new_window, gboolean reuse)
{
	GList *l = bi->list;
	BookMarkItem *child = NULL;

	while (l)
	{
		child = l->data;
		if (child->url)
			break;
		l = l->next;
	}

	if (l == NULL)
		return;

	/* reuse the first tab/window if required */
	/* FIXME: reusing the first tab/window often causes a crash, for
	   now a new window/tab is always created  --Josh */
//	if (reuse)
//		embed_load_url (embed, child->url);
//	else
//		embed = embed_create_from_url (embed, child->url, TRUE);

	if (reuse)
		embed = embed_create_from_url (embed, child->url, new_window);
	else
		embed = embed_create_from_url (embed, child->url, TRUE);	

	/* process the remaining bookmarks */
	l = l->next;
	while (l)
	{
		BookMarkItem *child = l->data;
		if (child->url)
			embed_create_from_url (embed, child->url, new_window);
		l = l->next;
	}
}

/* FIXME: Unused? */
/**
 * bookmarks_add_completion_recursively: recursively add bookmarks to the
 * auto completions table
 */
static void bookmarks_add_completion_recursively (BookMarkItem *b)
{
	GList *item;
  
	switch (b->type)
	{
	case CATEGORY:
	case AUTOBOOKMARKS:
		for (item = b->list; item != NULL; item = item->next)
		{
			bookmarks_add_completion_recursively (item->data);
		}
		break;

	case SITE:
		auto_completion_add_url (b->url);
		break;

	case SEPARATOR:
		break;
	}

}

/**
 * bookmarks_get_siteicon: Parse an http URL ans return the appropriate icon 
 */
GdkPixbuf *
bookmarks_get_siteicon (gchar *url, gboolean fetch)
{
	GnomeVFSURI *vfs_uri;
	gchar *favicon_hostname, *favicon_path;
	GdkPixbuf *result;

	if (!gnome_config_get_bool (CONF_GENERAL_FAVICONS_ENABLED))
	{
		return gdk_pixbuf_copy (site_pixmap);
	}

	if ((url == NULL) || (strncmp(url, "http", 4) != 0))
	{
		return gdk_pixbuf_copy (site_pixmap);
	}
	
	vfs_uri = gnome_vfs_uri_new(url);
	if (vfs_uri == NULL)
	{
		return gdk_pixbuf_copy (site_pixmap);
	} 
	
	favicon_hostname = g_strdup (gnome_vfs_uri_get_host_name (vfs_uri));
	if (favicon_hostname == NULL)
	{
		return gdk_pixbuf_copy (site_pixmap);
	} 

	gnome_vfs_uri_unref (vfs_uri);
	
	favicon_path = g_strconcat(g_get_home_dir (), "/.galeon/favicons/", 
				   favicon_hostname, "_favicon.ico", NULL);
	
	if (g_file_exists (favicon_path) && !fetch)
	{
		result = gdk_pixbuf_new_from_file (favicon_path);
	}
	else if (fetch)
	{
		if (get_siteicon(url))
		{
			result = gdk_pixbuf_new_from_file (favicon_path);
		}
		else
		{
			result = gdk_pixbuf_copy (site_pixmap);
		}
	}
	else
	{
		result = gdk_pixbuf_copy (site_pixmap);
	}

	g_free (favicon_path);
	g_free (favicon_hostname);
	
	if (gdk_pixbuf_get_width (result) > 16 || 
	    gdk_pixbuf_get_height (result) > 16)
	{
		GdkPixbuf *scaled_icon = gdk_pixbuf_scale_simple (result, 
						16, 16, GDK_INTERP_NEAREST);
		gdk_pixbuf_unref (result);
		return scaled_icon;
	}
	
	return result;
}
